#ifndef PATHMANAGER_HPP
#define PATHMANAGER_HPP
#include <iostream>
#include <cmath>
#include "Vec2.hpp"
#include <vector>
#include <boost/shared_ptr.hpp>
#include <exception>
#include "Utilities.hpp"
// #include "lua/llex.h"
#include <set>

struct PointData{
  Vec2 pos;
  int segment;
  PointData(Vec2 pos , int seg):pos(pos),segment(seg)
  {  
  }
};

struct Point{
  Vec2 pos;
  bool selected;
  Point(Vec2 pos):
    pos(pos),
    selected(false)
  {
    
  }
  Point(int x, int y):
    pos(x,y),
    selected(false)
  {
  }
};

///Path class wraps around a vector of points ensuring that lookup NEVER fails.
class Path{
public:
    typedef std::vector<Point> PointVector;
private:
    PointVector m_points;
    bool m_selected;
public:
    Path(){
	DebugStream << "Path constructed"<<std::endl;
    }
    Point operator[] (int index){
	return m_points[index % m_points.size()];
    }
    Point& at (int index){
	return m_points.at(index %m_points.size());
    }
    void move(Vec2 delta){
	for(int n = 0; n < m_points.size();++n){
	    m_points.at(n).pos +=delta;
	}
      
    }
    
    void add_point(Point point){
	m_points.push_back(point);
    }
    void add_point(int x,int y){
	add_point(Point(x,y));
    }
    void set_point(int index,Vec2 value){
	if(index > m_points.size()-1)return;
	m_points.at(index) = Point(value);
	
    }
    int size(){
	  return m_points.size();
    }
    PointData lin_interp(int activeSeg,float& val){
	//check to see if value is over 1...
	if(val >1.0f){
	 double fract,inte;
	 fract = modf(val,&inte);
// 	 DebugStream << "Val over 1 << resolving"<<std::endl;
// 	 DebugStream << "fract "<< fract<<std::endl;
	 val = fract;
// 	 activeSeg +=1;
	}
// 	DebugStream << "interpolating with val "<< val<<std::endl;
      
      
      //ensure active seg is in the right range..
	if(activeSeg >= m_points.size()-2){
// 	    DebugStream << "lin overflow"<<std::endl;
	    activeSeg = activeSeg % (m_points.size()-1);
	}
// 	DebugStream << "seg = "<< activeSeg<<std::endl;
// 	DebugStream << "lower index "<< activeSeg  << " upper "<< activeSeg + 1 <<std::endl;
	Point lower = at(activeSeg);
	Point upper = at(activeSeg+1);
	Point pos = ((1.0f - val) * lower.pos )+ (val * upper.pos);  
	if(val >=1 ){
	    activeSeg +=1;
	}
	return PointData(pos.pos,activeSeg);
	
    }
    void remove_point(int index){
	if(index > m_points.size() -1)return;
	PointVector::iterator it = m_points.begin() + index;
	m_points.erase(it);
    }
    bool empty(){
      return m_points.empty();
    }
    void remove_all(){
      m_points.clear();
    }
    
};

inline std::ostream& operator << (std::ostream& s, Path path){
      s << "Path:\n";
      for(int n = 0; n< path.size();++n ){
	  s << "\t"<< path[n].pos<<"\n";
      }
      return s;
}

typedef boost::shared_ptr<Path> PathPtr;
typedef std::vector<PathPtr> PathVector;

class PathLoadException : public std::exception{
  const char* m_msg;
public:
  PathLoadException(const char* msg):m_msg(msg){
  }
  const char* what() const throw(){
    return m_msg;
  }
};




class SelectedPointObject{
public:
  typedef std::set<Point*> PointSet;
private:
  PointSet m_pointSet;
  Point* m_primarySelection;
public:
  SelectedPointObject();
  virtual ~SelectedPointObject();
  void add_to_selection(Point* point);
  void remove_selection(Point* point);
  void set_primary_selection(Point* point);
  void clear_selection();
  void clear_primary();
  Point* get_primary();
  PointSet& get_node_set();
  PointSet::iterator begin();
  PointSet::iterator end();
  bool empty();
  int selection_size();
  
};












class PathManager
{
  PathManager();
  PathVector m_paths;
  SelectedPointObject m_selectedPointObject;
public:
  static PathManager& get_instance();
  int get_paths_size();
  PathPtr get_path(int index);
  int add_path(PathPtr path);
  void remove_path(int index);
  void move_path(int index, Vec2 delta);
  //load and save
  void save_path_database(std::string filepath);
  void load_path_database(std::string filepath);
  
  SelectedPointObject& get_selected_point_objects();
  
};

#endif // PATHMANAGER_HPP
