#ifndef TREENODE_HPP
#define TREENODE_HPP
#include <string>
#include <map>
#include <vector>
#include <ostream>
#include <boost/variant.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <cassert>
#include "MotionModel.hpp"
#include <exception>
#include <boost/bind.hpp>
#include <boost/function.hpp>

typedef boost::variant<int,bool,float,std::string, Vec2 > PropertyValue;
typedef boost::function<void (int)> Trigger;
typedef unsigned int PropertyFlags;

static const int SC_OFFSET = 1000;

typedef std::vector<std::string> StringVector;

class Property{
  public:
    enum Flags{
	PF_NONE = 0,
	PF_READ = 1 << 0,
	PF_WRITE = 1 << 1,
	PF_SEND = 1 << 2
    };
  private:
    PropertyFlags m_flags; 
    PropertyValue m_value;
    std::string m_oscPath;
  public:
    class Exception : public std::exception{
      const char* m_msg;
      public:
	Exception(const char* msg) : m_msg(msg) {}
        virtual const char* what() const throw() { return m_msg; }
    };
    
    Property(PropertyValue value,PropertyFlags flags=Property::PF_READ | PF_WRITE):m_value(value),m_flags(flags){
    };
   
    const PropertyValue& get_value()const{
	if(m_flags & PF_READ){
	   return m_value;
	} else {
	  throw Property::Exception("No read permission on property.");
	}
    }
    void set_value(const PropertyValue& value){
	if(m_flags & PF_WRITE){
	    m_value = value;
	}else{
	   throw Property::Exception("No write permission on property.");
	}
    }
    PropertyFlags get_flags()const{
	return m_flags;
    }
    void set_flags(PropertyFlags flags){
	if(m_flags & PF_WRITE){
	    m_flags = flags;
	}else{
	   throw Property::Exception("No write permission on property when setting flags.");
	}
    }
    std::string& get_osc_path(){
	return m_oscPath;
    }
    void set_osc_path(const std::string& path){
	m_oscPath = path;
    }
};

typedef std::map<std::string,Property > PropertyMap;

class TreeNode : public boost::enable_shared_from_this<TreeNode>

{
  public:
class IdException : public std::exception{
  public:
	const int id;
	IdException(const int _id):id(_id){}
        virtual const char* what() const throw(){ return "Node Id was invalid" ;}
  };
  class ChildIdException : public std::exception{
  public:
	const int id;
	ChildIdException(const int _id):id(_id){}
        virtual const char* what() const throw(){ return "Child Id was invalid" ;}
  };  
  class NodeException : public std::exception{
  public:
	TreeNode* node;
	NodeException( TreeNode* node):node(node){}
        virtual const char* what() const throw(){ return "Child Id was invalid" ;}
  };  

  
  typedef boost::shared_ptr<TreeNode> Ptr;
  typedef boost::weak_ptr<TreeNode> WPtr;
  typedef std::vector<Ptr> PtrVector;
  typedef std::map<int,WPtr > TreeNodeMap;
  enum TreeNodeType{
      TN_TREE=0,
      TN_AGENT,
      TN_WORLD
  };
private:
  PropertyMap m_propertyMap;
  PtrVector m_children,m_pendingChildren;
  int m_id,m_scId;
  static int g_id; 
  static TreeNodeMap g_nodeMap;
  WPtr m_parent;
  bool m_isAlive;
  bool m_isSelected;
  bool m_hasChildren;
  void update_pending_children();
  
  
  bool m_active;
  StringVector m_errors;
  std::string m_scriptName;  
  float m_lifeTime,m_activeTime;
  MotionModel::Ptr m_motionModel;
  void initialize_sc_connection();
  void kill_sc_node();
 protected:
  TreeNode(Vec2 pos);

public:
  static Trigger s_constructionTrigger;
  static Trigger s_destructionTrigger;

  virtual ~TreeNode();
  
  //properties
  const Property& get_property(const std::string& name)const;
  Property& get_property(const std::string& name);
  void add_property(const std::string& name, const Property& value);
  void set_property(const std::string& name ,const PropertyValue& value);
  void set_property_osc_path(const std::string& name ,const std::string& path);
  void print_properties(std::ostream& stream)const;
  const int get_property_count()const;
  PropertyMap::const_iterator get_property_at(int index)const;
  PropertyMap::iterator get_property_at(int index);
  void copy_properties_from(const TreeNode& node);
  bool property_exists(const std::string& name);
  
  //id
  int get_id()const;
  
  //children
  Ptr add_child(Ptr child);
  TreeNode& get_child_at(int index);
  const TreeNode& get_child_at(int index)const;
  void remove_child_at(int index);
  const int number_of_children()const ;
  PtrVector::iterator get_child_begin();
  PtrVector::const_iterator get_child_begin()const;
  PtrVector::iterator get_child_end();
  PtrVector::const_iterator get_child_end()const;
  
  
  void kill_children();
  bool has_children();
  
  //parent
  WPtr get_parent();
  const WPtr get_parent()const;
  int get_parent_id();
  
  //show functions
  static void show_node_hierarchy(std::ostream& stream,const TreeNode::Ptr& node, int depth=0);
  static void show_node_usage(std::ostream& stream);

  const Vec2& get_pos()const;
  void set_pos(Vec2 pos);
  
  //get any node
  static WPtr get_treenode_by_id(int id);
  
  
  virtual void kill();
  bool is_alive()const;
  
  //animation function
  virtual void tick(double dT);
  
// DEPRECATED  virtual void compute_grid_attraction(double dT);
  
  void remove_treenode_from_existence();
  
  static TreeNode::Ptr factory(TreeNodeType type,Vec2 pos ,const char* brain="");
  ///Returns the type of treenode
  virtual TreeNodeType is_type()const;
  
  static TreeNodeMap::const_iterator get_global_node_list_begin();
  static TreeNodeMap::const_iterator get_global_node_list_end();
  
  float get_lifeTime()const;
  
  void set_motion_model(MotionModel::MotionType type);
  void set_selection_state(bool state);
  bool is_selected();  
  bool is_within(Vec2 pos);//function for mouse detection
  void move(Vec2 v);
  
  void set_active_state(bool state);
  bool get_activity_state();
  
  void add_error(std::string msg);
  StringVector& get_errors();
  void clear_errors();
  
  
  void set_script_name(std::string name);
  std::string get_script_name();
};

int find_nearest_node(Vec2 pos);

typedef std::vector<int > IdVector;

static IdVector s_nearbyIDVector; 

IdVector* get_nearby_ids(Vec2 pos, int radius);


#endif // TREENODE_HPP
